@@ -139,7 +139,8 @@ let rec arg_deps st ?ignore params args =
       | Some y' when Var.equal y y' -> ()
       | _ -> add_assign_def st x y);
       arg_deps st params args
-  | _ -> ()
+  | [], [] -> ()
+  | _ -> assert false
 
 let cont_deps blocks st ?ignore (pc, args) =
   let block = Addr.Map.find pc blocks in
@@ -153,6 +154,7 @@ let expr_deps blocks st x e =
   match e with
   | Constant _ | Prim ((Vectlength | Not | IsInt | Eq | Neq | Lt | Le | Ult), _) | Block _
     -> ()
+  | Special _ -> ()
   | Prim ((Extern ("caml_check_bound" | "caml_array_unsafe_get") | Array_get), l) ->
       (* The analysis knowns about these primitives, and will compute
          an approximation of the value they return based on an
@@ -215,7 +217,9 @@ let expr_deps blocks st x e =
       match st.defs.(Var.idx f) with
       | Expr (Closure (params, _)) when List.length args = List.length params ->
           Hashtbl.add st.applied_functions (x, f) ();
-          if not st.fast then List.iter2 ~f:(fun p a -> add_assign_def st p a) params args;
+          if st.fast
+          then List.iter ~f:(fun a -> do_escape st Escape a) args
+          else List.iter2 ~f:(fun p a -> add_assign_def st p a) params args;
           Var.Set.iter (fun y -> add_dep st x y) (Var.Map.find f st.return_values)
       | _ -> ())
   | Closure (l, cont) ->
@@ -243,28 +247,35 @@ let program_deps st { blocks; _ } =
       | Cond (x, cont1, cont2) ->
           cont_deps blocks st cont1;
           cont_deps blocks st ~ignore:x cont2
-      | Switch (x, a1, a2) ->
+      | Switch (x, a1) -> (
           Array.iter a1 ~f:(fun cont -> cont_deps blocks st cont);
-          Array.iter a2 ~f:(fun cont -> cont_deps blocks st cont);
-          let h = Hashtbl.create 16 in
-          Array.iteri
-            ~f:(fun i (pc, _) ->
-              Hashtbl.replace h pc (i :: (try Hashtbl.find h pc with Not_found -> [])))
-            a2;
           if not st.fast
           then
-            Hashtbl.iter
-              (fun pc tags ->
-                let block = Addr.Map.find pc blocks in
-                List.iter
-                  ~f:(fun (i, _) ->
-                    match i with
-                    | Let (y, Field (x', _)) when Var.equal x x' ->
-                        Hashtbl.add st.known_cases y tags
-                    | _ -> ())
-                  block.body)
-              h
-      | Pushtrap (cont, x, cont_h, _) ->
+            (* looking up the def of x is fine here, because the tag
+               we're looking for is at addr [pc - 2] (see
+               parse_bytecode.ml) and [Addr.Map.iter] iterate in
+               increasing order *)
+            match st.defs.(Code.Var.idx x) with
+            | Expr (Prim (Extern "%direct_obj_tag", [ Pv b ])) ->
+                let h = Hashtbl.create 16 in
+                Array.iteri a1 ~f:(fun i (pc, _) ->
+                    Hashtbl.replace
+                      h
+                      pc
+                      (i :: (try Hashtbl.find h pc with Not_found -> [])));
+                Hashtbl.iter
+                  (fun pc tags ->
+                    let block = Addr.Map.find pc blocks in
+                    List.iter
+                      ~f:(fun (i, _) ->
+                        match i with
+                        | Let (y, Field (x', _)) when Var.equal b x' ->
+                            Hashtbl.add st.known_cases y tags
+                        | _ -> ())
+                      block.body)
+                  h
+            | Expr _ | Phi _ -> ())
+      | Pushtrap (cont, x, cont_h) ->
           add_var st x;
           st.defs.(Var.idx x) <- Phi { known = Var.Set.empty; others = true };
           cont_deps blocks st cont_h;
@@ -314,12 +325,13 @@ module Domain = struct
     then (
       st.may_escape.(idx) <- s;
       match st.defs.(idx) with
-      | Expr (Block (_, a, _)) ->
+      | Expr (Block (_, a, _)) -> (
           Array.iter ~f:(fun y -> variable_escape ~update ~st ~approx s y) a;
-          if Poly.equal s Escape
-          then (
-            st.possibly_mutable.(idx) <- true;
-            update ~children:true x)
+          match s with
+          | Escape ->
+              st.possibly_mutable.(idx) <- true;
+              update ~children:true x
+          | Escape_constant | No -> ())
       | Expr (Closure (params, _)) ->
           List.iter
             ~f:(fun y ->
@@ -450,6 +462,7 @@ let propagate st ~update approx x =
              block *)
           Domain.bot
       | Prim (Extern _, _) -> Domain.others
+      | Special _ -> Domain.others
       | Apply { f; args; _ } -> (
           match Var.Tbl.get approx f with
           | Values { known; others } ->
@@ -470,8 +483,13 @@ let propagate st ~update approx x =
                       if not (Hashtbl.mem st.applied_functions (x, g))
                       then (
                         Hashtbl.add st.applied_functions (x, g) ();
-                        if not st.fast
+                        if st.fast
                         then
+                          List.iter
+                            ~f:(fun y ->
+                              Domain.variable_escape ~update ~st ~approx Escape y)
+                            args
+                        else
                           List.iter2
                             ~f:(fun p a ->
                               add_assign_def st p a;
@@ -540,6 +558,8 @@ type info =
   { info_defs : def array
   ; info_approximation : Domain.t Var.Tbl.t
   ; info_may_escape : Var.ISet.t
+  ; info_variable_may_escape : escape_status array
+  ; info_return_vals : Var.Set.t Var.Map.t
   }
 
 let f ~fast p =
@@ -593,7 +613,7 @@ let f ~fast p =
               | Values { known; others } ->
                   Format.fprintf
                     f
-                    "{%a/%b} mut:%b vmut:%b esc:%s"
+                    "{%a/%b} mut:%b vmut:%b vesc:%s esc:%s"
                     (Format.pp_print_list
                        ~pp_sep:(fun f () -> Format.fprintf f ", ")
                        (fun f x ->
@@ -604,28 +624,38 @@ let f ~fast p =
                            x
                            (match st.defs.(Var.idx x) with
                            | Expr (Closure _) -> "C"
-                           | Expr (Block _) ->
+                           | Expr (Block _) -> (
                                "B"
                                ^
-                               if Poly.equal st.may_escape.(Var.idx x) Escape
-                               then "X"
-                               else ""
+                               match st.may_escape.(Var.idx x) with
+                               | Escape -> "X"
+                               | _ -> "")
                            | _ -> "O")))
                     (Var.Set.elements known)
                     others
                     st.possibly_mutable.(Var.idx x)
                     st.variable_possibly_mutable.(Var.idx x)
+                    (match st.variable_may_escape.(Var.idx x) with
+                    | Escape -> "Y"
+                    | Escape_constant -> "y"
+                    | No -> "n")
                     (match st.may_escape.(Var.idx x) with
                     | Escape -> "Y"
                     | Escape_constant -> "y"
                     | No -> "n"))
             s)
       vars;
+  let info_variable_may_escape = variable_may_escape in
   let info_may_escape = Var.ISet.empty () in
   Array.iteri
     ~f:(fun i s -> if Poly.(s <> No) then Var.ISet.add info_may_escape (Var.of_idx i))
     may_escape;
-  { info_defs = defs; info_approximation = approximation; info_may_escape }
+  { info_defs = defs
+  ; info_approximation = approximation
+  ; info_variable_may_escape
+  ; info_may_escape
+  ; info_return_vals = rets
+  }
 
 let exact_call info f n =
   match Var.Tbl.get info.info_approximation f with
